/*
 * Copyright 2019 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.kafka.secretregistry.rest.resources;

import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.container.AsyncResponse;
import javax.ws.rs.container.Suspended;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.SecurityContext;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import io.confluent.kafka.secretregistry.client.rest.Versions;
import io.confluent.kafka.secretregistry.client.rest.entities.Secret;
import io.confluent.kafka.secretregistry.exceptions.SecretRegistryException;
import io.confluent.kafka.secretregistry.exceptions.SecretRegistryRequestForwardingException;
import io.confluent.kafka.secretregistry.exceptions.SecretRegistryStoreException;
import io.confluent.kafka.secretregistry.exceptions.SecretRegistryTimeoutException;
import io.confluent.kafka.secretregistry.exceptions.UnknownMasterException;
import io.confluent.kafka.secretregistry.rest.exceptions.Errors;
import io.confluent.kafka.secretregistry.storage.KafkaSecretRegistry;
import io.confluent.rest.annotations.PerformanceMetric;

@Path("/secret/paths/{path}/keys")
@Produces({Versions.SECRET_REGISTRY_V1_JSON_WEIGHTED,
    Versions.SECRET_REGISTRY_DEFAULT_JSON_WEIGHTED,
    Versions.JSON_WEIGHTED})
@Consumes({Versions.SECRET_REGISTRY_V1_JSON,
    Versions.SECRET_REGISTRY_DEFAULT_JSON,
    Versions.JSON, Versions.GENERIC_REQUEST})
public class PathKeyResource {

  private static final Logger log = LoggerFactory.getLogger(PathKeyResource.class);

  private final KafkaSecretRegistry secretRegistry;

  private final RequestHeaderBuilder requestHeaderBuilder = new RequestHeaderBuilder();

  public PathKeyResource(KafkaSecretRegistry registry) {
    this.secretRegistry = registry;
  }

  @GET
  @Path("/{key}")
  @PerformanceMetric("paths.keys.get-key")
  public List<Secret> getAllVersions(
      @Context SecurityContext securityContext,
      @PathParam("path") String path,
      @PathParam("key") String key
  ) {
    if (securityContext.getUserPrincipal() != null) {
      String name = securityContext.getUserPrincipal().getName();
      log.info("User '{}' getting versions for '{}:{}'", name, path, key);
    } else {
      log.info("Unknown user getting versions for '{}:{}'", path, key);
    }

    try {
      if (!secretRegistry.listPaths().contains(path)) {
        throw Errors.pathNotFoundException();
      }
      if (!secretRegistry.listKeys(path).contains(key)) {
        throw Errors.keyNotFoundException();
      }
      return secretRegistry.getAllVersions(path, key);
    } catch (SecretRegistryStoreException e) {
      throw Errors.storeException("Error while listing versions", e);
    } catch (SecretRegistryException e) {
      throw Errors.secretRegistryException("Error while listing versions", e);
    }
  }

  @GET
  @PerformanceMetric("paths.keys.list")
  public Set<String> list(
      @Context SecurityContext securityContext,
      @PathParam("path") String path
  ) {
    if (securityContext.getUserPrincipal() != null) {
      String name = securityContext.getUserPrincipal().getName();
      log.info("User '{}' listing keys for '{}'", name, path);
    } else {
      log.info("Unknown user listing keys for '{}'", path);
    }

    try {
      if (!secretRegistry.listPaths().contains(path)) {
        throw Errors.pathNotFoundException();
      }
      return secretRegistry.listKeys(path);
    } catch (SecretRegistryStoreException e) {
      throw Errors.storeException("Error while listing paths", e);
    } catch (SecretRegistryException e) {
      throw Errors.secretRegistryException("Error while listing paths", e);
    }
  }

  @DELETE
  @Path("/{key}")
  @PerformanceMetric("paths.keys.delete-key")
  public void deleteKey(
      final @Suspended AsyncResponse asyncResponse,
      @Context SecurityContext securityContext,
      @Context HttpHeaders headers,
      @PathParam("path") String path,
      @PathParam("key") String key
  ) {
    if (securityContext.getUserPrincipal() != null) {
      String name = securityContext.getUserPrincipal().getName();
      log.info("User '{}' deleting key '{}:{}'", name, path, key);
    } else {
      log.info("Unknown user deleting key '{}:{}'", path, key);
    }

    String errorMessage = "Error while validating that path "
        + path + " and key " + key
        + " exist in the registry";
    try {
      if (!secretRegistry.listPaths().contains(path)) {
        throw Errors.pathNotFoundException();
      }
      if (!secretRegistry.listKeys(path).contains(key)) {
        throw Errors.keyNotFoundException();
      }
    } catch (SecretRegistryStoreException e) {
      throw Errors.storeException(errorMessage, e);
    } catch (SecretRegistryException e) {
      throw Errors.secretRegistryException(errorMessage, e);
    }
    List<Integer> deletedVersions;
    try {
      Map<String, String> headerProperties = requestHeaderBuilder.buildRequestHeaders(
          headers, secretRegistry.config().whitelistHeaders());
      deletedVersions = secretRegistry.deleteKeyOrForward(headerProperties, path, key);
    } catch (SecretRegistryTimeoutException e) {
      throw Errors.operationTimeoutException("Delete Secret Version operation timed out", e);
    } catch (SecretRegistryStoreException e) {
      throw Errors.storeException("Delete Secret Version operation failed while writing"
          + " to the Kafka store", e);
    } catch (SecretRegistryRequestForwardingException e) {
      throw Errors
          .requestForwardingFailedException("Error while forwarding delete secret version request"
              + " to the master", e);
    } catch (UnknownMasterException e) {
      throw Errors.unknownMasterException("Master not known.", e);
    } catch (SecretRegistryException e) {
      throw Errors.secretRegistryException("Error while deleting Secret Version", e);
    }

    asyncResponse.resume(deletedVersions);
  }
}
